var __defProp = Object.defineProperty;
var __markAsModule = (target) => __defProp(target, "__esModule", { value: true });
var __export = (target, all) => {
  __markAsModule(target);
  for (var name in all)
    __defProp(target, name, { get: all[name], enumerable: true });
};

// src/geometries/buffer-geometry.ts
var BufferAttribute = class {
  constructor(data, recordSize) {
    this.data = data;
    this.recordSize = recordSize;
  }
  get count() {
    return this.data.length / this.recordSize;
  }
};
var BufferGeometry = class {
  constructor() {
    this.attributes = {};
    this.count = 0;
    this.index = null;
    this.indexType = 0;
    this.groups = [];
  }
  setIndex(indices, bits = 16) {
    this.index = Array.from(indices);
    this.indexType = bits;
    this.count = this.index.length;
    return this;
  }
  setAttribute(attributeName, bufferAttribute) {
    this.attributes[attributeName] = bufferAttribute;
    if (!this.index) {
      this.count = Math.max(this.count, bufferAttribute.count);
    }
    return this;
  }
  removeAttribute(attributeName) {
    if (attributeName in this.attributes) {
      delete this.attributes[attributeName];
    }
    return this;
  }
  dispose() {
    for (const attribute of Object.keys(this.attributes)) {
      this.removeAttribute(attribute);
    }
  }
};

// src/geometries/box.ts
function buildPlane(u, v, w, udir, vdir, width, height, depth, gridX, gridY, vertexIndexStart, faceIndex) {
  const segmentWidth = width / gridX;
  const segmentHeight = height / gridY;
  const widthHalf = width / 2;
  const heightHalf = height / 2;
  const depthHalf = depth / 2;
  const gridX1 = gridX + 1;
  const gridY1 = gridY + 1;
  let vertexCount = 0;
  let groupCount = 0;
  const vector = { x: NaN, y: NaN, z: NaN };
  const vertices = [];
  const normals = [];
  const uvs = [];
  const indices = [];
  const faceIndices = [];
  for (let iy = 0; iy < gridY1; iy++) {
    const y = iy * segmentHeight - heightHalf;
    for (let ix = 0; ix < gridX1; ix++) {
      const x = ix * segmentWidth - widthHalf;
      vector[u] = x * udir;
      vector[v] = y * vdir;
      vector[w] = depthHalf;
      vertices.push(vector.x, vector.y, vector.z);
      vector[u] = 0;
      vector[v] = 0;
      vector[w] = depth > 0 ? 1 : -1;
      normals.push(vector.x, vector.y, vector.z);
      uvs.push(ix / gridX);
      uvs.push(1 - iy / gridY);
      faceIndices.push(faceIndex);
      vertexCount += 1;
    }
  }
  for (let iy = 0; iy < gridY; iy++) {
    for (let ix = 0; ix < gridX; ix++) {
      const a = vertexIndexStart + ix + gridX1 * iy;
      const b = vertexIndexStart + ix + gridX1 * (iy + 1);
      const c = vertexIndexStart + (ix + 1) + gridX1 * (iy + 1);
      const d = vertexIndexStart + (ix + 1) + gridX1 * iy;
      indices.push(a, b, d);
      indices.push(b, c, d);
      groupCount += 6;
    }
  }
  return {
    vertices,
    normals,
    uvs,
    indices,
    faceIndices,
    groupCount,
    vertexCount
  };
}
function createBoxGeometry(width = 1, height = 1, depth = 1, widthSegments = 1, heightSegments = 1, depthSegments = 1) {
  let count = 0;
  const px = buildPlane("z", "y", "x", -1, -1, depth, height, width, depthSegments, heightSegments, count, 0);
  count += px.vertexCount;
  const nx = buildPlane("z", "y", "x", 1, -1, depth, height, -width, depthSegments, heightSegments, count, 1);
  count += nx.vertexCount;
  const py = buildPlane("x", "z", "y", 1, 1, width, depth, height, widthSegments, depthSegments, count, 2);
  count += py.vertexCount;
  const ny = buildPlane("x", "z", "y", 1, -1, width, depth, -height, widthSegments, depthSegments, count, 3);
  count += ny.vertexCount;
  const pz = buildPlane("x", "y", "z", 1, -1, width, height, depth, widthSegments, heightSegments, count, 4);
  count += pz.vertexCount;
  const nz = buildPlane("x", "y", "z", -1, -1, width, height, -depth, widthSegments, heightSegments, count, 5);
  count += nz.vertexCount;
  const vertices = new Float32Array(Array.prototype.concat.apply([], [
    px.vertices,
    nx.vertices,
    py.vertices,
    ny.vertices,
    pz.vertices,
    nz.vertices
  ]));
  const normals = new Float32Array(Array.prototype.concat.apply([], [px.normals, nx.normals, py.normals, ny.normals, pz.normals, nz.normals]));
  const uvs = new Float32Array(Array.prototype.concat.apply([], [px.uvs, nx.uvs, py.uvs, ny.uvs, pz.uvs, nz.uvs]));
  const indices = Array.prototype.concat.apply([], [px.indices, nx.indices, py.indices, ny.indices, pz.indices, nz.indices]);
  const faceIndices = new Float32Array(Array.prototype.concat.apply([], [
    px.faceIndices,
    nx.faceIndices,
    py.faceIndices,
    ny.faceIndices,
    pz.faceIndices,
    nz.faceIndices
  ]));
  const geometry = new BufferGeometry();
  let groupIndex = 0;
  geometry.groups.push({ startIndex: groupIndex, count: px.indices.length });
  groupIndex += px.indices.length;
  geometry.groups.push({ startIndex: groupIndex, count: nx.indices.length });
  groupIndex += nx.indices.length;
  geometry.groups.push({ startIndex: groupIndex, count: py.indices.length });
  groupIndex += py.indices.length;
  geometry.groups.push({ startIndex: groupIndex, count: ny.indices.length });
  groupIndex += ny.indices.length;
  geometry.groups.push({ startIndex: groupIndex, count: pz.indices.length });
  groupIndex += pz.indices.length;
  geometry.groups.push({ startIndex: groupIndex, count: nz.indices.length });
  geometry.setAttribute("position", new BufferAttribute(vertices, 3));
  geometry.setAttribute("normal", new BufferAttribute(normals, 3));
  geometry.setAttribute("uv", new BufferAttribute(uvs, 2));
  geometry.setAttribute("faceIndex", new BufferAttribute(faceIndices, 1));
  geometry.setIndex(indices);
  return geometry;
}

// src/utils/color.ts
var Color = class {
  constructor(red, green, blue, alpha = 255) {
    this.red = red;
    this.green = green;
    this.blue = blue;
    this.alpha = alpha;
  }
  static fromHex(color) {
    return new Color(parseInt(color.slice(1, 3), 16), parseInt(color.slice(3, 5), 16), parseInt(color.slice(5, 7), 16), color.length === 9 ? parseInt(color.slice(7, 9), 16) : 255);
  }
  toVec3() {
    const { red, green, blue } = this;
    return [red / 255, green / 255, blue / 255];
  }
  toVec4() {
    const { red, green, blue, alpha } = this;
    return [red / 255, green / 255, blue / 255, alpha / 255];
  }
};

// src/utils/faces-to-buffer.ts
function facesToBuffer(faces, vertices) {
  return faces.map((face) => {
    if (face.length === 3) {
      return face.map((vertexIndex) => vertices[vertexIndex]);
    }
    if (face.length === 4) {
      const q = face.map((vertexIndex) => vertices[vertexIndex]);
      return [q[0], q[1], q[3], q[3], q[1], q[2]];
    }
    throw Error("not supported");
  }).flat().map((v) => v.toArray()).flat();
}

// src/utils/surface-normals.ts
function calculateSurfaceNormal(p1, p2, p3) {
  const u = p2.sub(p1);
  const v = p3.sub(p1);
  return u.cross(v).normalized;
}

// src/utils/vector.ts
var Vector = class {
  constructor(...values) {
    this.values = values;
  }
  get dim() {
    return this.values.length;
  }
  get x() {
    return this.values[0];
  }
  set x(value) {
    this.values[0] = value;
  }
  get y() {
    return this.values[1];
  }
  set y(value) {
    this.values[1] = value;
  }
  get z() {
    return this.values[2];
  }
  set z(value) {
    this.values[2] = value;
  }
  get w() {
    return this.values[3];
  }
  set w(value) {
    this.values[3] = value;
  }
  get xy() {
    return new Vector(this.values[0], this.values[1]);
  }
  get xz() {
    return new Vector(this.values[0], this.values[2]);
  }
  get yz() {
    return new Vector(this.values[1], this.values[2]);
  }
  get xyz() {
    return new Vector(this.values[0], this.values[1], this.values[2]);
  }
  set(...values) {
    this.values.splice(0, this.values.length, ...values);
    return this;
  }
  static fromArray(arr) {
    return new Vector(...arr);
  }
  static fromNumber(n, dim) {
    return new Vector(...Array(dim).fill(n));
  }
  clone() {
    return new Vector(...this.values);
  }
  add(otherVec) {
    return new Vector(...this.values.map((v, idx) => v + otherVec.values[idx]));
  }
  sub(otherVec) {
    return new Vector(...this.values.map((v, idx) => v - otherVec.values[idx]));
  }
  mul(value) {
    return new Vector(...this.values.map((v) => v * value));
  }
  div(value) {
    return new Vector(...this.values.map((x) => x / value));
  }
  dot(otherVec) {
    return this.values.map((x, idx) => x * otherVec.values[idx]).reduce((a, b) => a + b);
  }
  equals(otherVec) {
    return this.values.map((v, idx) => v === otherVec.values[idx]).reduce((a, b) => a === b);
  }
  get length() {
    return Math.sqrt(this.values.map((v) => v ** 2).reduce((a, b) => a + b));
  }
  toArray() {
    return this.values.slice(0);
  }
  toString() {
    return `(${this.values.join(", ")})`;
  }
  cross(otherVec) {
    if (this.dim !== 3 || otherVec.dim !== 3) {
      throw Error("dimension not supported");
    }
    return new Vector(this.y * otherVec.z - this.z * otherVec.y, this.z * otherVec.x - this.x * otherVec.z, this.x * otherVec.y - this.y * otherVec.x);
  }
  get normalized() {
    return this.div(this.length);
  }
};

// src/utils/matrix.ts
var Matrix = class {
  constructor(values, options) {
    this.values = values;
    if (options) {
      this.numRows = options.numRows;
      this.numCols = options.numCols;
    } else {
      const dimension = Math.sqrt(values.length);
      if (Number.isInteger(dimension)) {
        this.numCols = this.numRows = dimension;
        return;
      }
      throw Error("ArgumentError");
    }
  }
  static identity(dimension) {
    if (dimension <= 0 || !Number.isInteger(dimension)) {
      throw Error("ArgumentError");
    }
    return new Matrix(Array(dimension ** 2).fill(0).map((_, i) => i % dimension === (i / dimension | 0) ? 1 : 0));
  }
  static fromVector(value) {
    if (value instanceof Vector) {
      return new Matrix(value.toArray(), { numRows: value.dim, numCols: 1 });
    }
    throw Error("unsupported type");
  }
  static fromVectors(vectors) {
    if (!vectors || vectors.length === 0) {
      throw Error("Argument error.");
    }
    const dimensions = vectors.map((v) => v.dim);
    const dimensionsMatch = dimensions.every((x) => x === dimensions[0]);
    const dimension = dimensions[0];
    if (!dimensionsMatch) {
      throw Error("Dimensions mismatch.");
    }
    const matrix = Array(dimension * vectors.length);
    for (let i = 0; i < vectors.length; i++) {
      for (let j = 0; j < dimension; j++) {
        matrix[i + j * vectors.length] = vectors[i].values[j];
      }
    }
    return new Matrix(matrix, { numRows: dimension, numCols: vectors.length });
  }
  toArray() {
    return this.values;
  }
  valueAt(row, column) {
    return this.values[column * this.numRows + row];
  }
  colAt(column) {
    const { numRows } = this;
    return this.values.slice(column * numRows, column * numRows + numRows);
  }
  rowAt(row) {
    const { numRows, numCols } = this;
    return Array(numCols).fill(0).map((_, column) => this.values[column * numRows + row]);
  }
  transpose() {
    const transposedValues = [];
    Array(this.numRows).fill(0).map((_, i) => {
      transposedValues.push(...this.rowAt(i));
    });
    return new Matrix(transposedValues, {
      numRows: this.numCols,
      numCols: this.numRows
    });
  }
  equals(otherMatrix) {
    if (this.values.length !== otherMatrix.values.length || this.numCols !== otherMatrix.numCols || this.numRows !== otherMatrix.numRows) {
      return false;
    }
    for (let i = 0; i < this.values.length; i++) {
      if (this.values[i] !== otherMatrix.values[i]) {
        return false;
      }
    }
    return true;
  }
  roughlyEquals(otherMatrix, delta = 1e-14) {
    if (this.values.length !== otherMatrix.values.length || this.numCols !== otherMatrix.numCols || this.numRows !== otherMatrix.numRows) {
      return false;
    }
    for (let i = 0; i < this.values.length; i++) {
      if (Math.abs(this.values[i] - otherMatrix.values[i]) > delta) {
        return false;
      }
    }
    return true;
  }
  add(otherMatrix) {
    if (this.numCols === otherMatrix.numCols && this.numRows === otherMatrix.numRows && this.values.length === otherMatrix.values.length) {
      const newValues = this.values.map((value, i) => value + otherMatrix.values[i]);
      return new Matrix(newValues, {
        numRows: this.numRows,
        numCols: this.numCols
      });
    }
    throw Error("ArgumentError");
  }
  sub(otherMatrix) {
    if (this.numCols === otherMatrix.numCols && this.numRows === otherMatrix.numRows && this.values.length === otherMatrix.values.length) {
      const newValues = this.values.map((value, i) => value - otherMatrix.values[i]);
      return new Matrix(newValues, {
        numRows: this.numRows,
        numCols: this.numCols
      });
    }
    throw Error("ArgumentError");
  }
  mul(param) {
    if (typeof param === "number") {
      const multipliedValues = this.values.map((value) => value * param);
      return new Matrix(multipliedValues, {
        numRows: this.numRows,
        numCols: this.numCols
      });
    }
    if (param instanceof Vector) {
      if (param.dim !== this.numCols) {
        throw Error("dimension mismatch");
      }
      const m = this.mul(Matrix.fromVector(param));
      return new Vector(...m.values);
    }
    if (param instanceof Matrix) {
      const mat = param;
      const { numRows } = this;
      const { numCols } = mat;
      const multipliedValues = Array(numRows * numCols).fill(0).map((_, idx) => {
        const y = idx % numRows;
        const x = idx / numRows | 0;
        const row = this.rowAt(y);
        const col = mat.colAt(x);
        return row.map((value, i) => value * col[i]).reduce((a, b) => a + b);
      });
      return new Matrix(multipliedValues, { numRows, numCols });
    }
    throw Error("ArgumentError");
  }
  div(param) {
    if (typeof param === "number") {
      const multipliedValues = this.values.map((value) => value / param);
      return new Matrix(multipliedValues, {
        numRows: this.numRows,
        numCols: this.numCols
      });
    }
    if (param instanceof Matrix) {
      const mat = param;
      const inversed = mat.inverse();
      return this.mul(inversed);
    }
    throw Error("ArgumentError");
  }
  determinant() {
    const { numRows, numCols } = this;
    const m = (col, row) => this.valueAt(row, col);
    if (numRows !== numCols) {
      throw Error("ArgumentError");
    }
    if (numRows === 2) {
      return m(0, 0) * m(1, 1) - m(0, 1) * m(1, 0);
    }
    if (numRows === 3) {
      const a00 = m(0, 0), a01 = m(0, 1), a02 = m(0, 2);
      const a10 = m(1, 0), a11 = m(1, 1), a12 = m(1, 2);
      const a20 = m(2, 0), a21 = m(2, 1), a22 = m(2, 2);
      const b01 = a22 * a11 - a12 * a21;
      const b11 = -a22 * a10 + a12 * a20;
      const b21 = a21 * a10 - a11 * a20;
      return a00 * b01 + a01 * b11 + a02 * b21;
    }
    if (numRows === 4) {
      const a00 = m(0, 0), a01 = m(0, 1), a02 = m(0, 2), a03 = m(0, 3);
      const a10 = m(1, 0), a11 = m(1, 1), a12 = m(1, 2), a13 = m(1, 3);
      const a20 = m(2, 0), a21 = m(2, 1), a22 = m(2, 2), a23 = m(2, 3);
      const a30 = m(3, 0), a31 = m(3, 1), a32 = m(3, 2), a33 = m(3, 3);
      const b00 = a00 * a11 - a01 * a10;
      const b01 = a00 * a12 - a02 * a10;
      const b02 = a00 * a13 - a03 * a10;
      const b03 = a01 * a12 - a02 * a11;
      const b04 = a01 * a13 - a03 * a11;
      const b05 = a02 * a13 - a03 * a12;
      const b06 = a20 * a31 - a21 * a30;
      const b07 = a20 * a32 - a22 * a30;
      const b08 = a20 * a33 - a23 * a30;
      const b09 = a21 * a32 - a22 * a31;
      const b10 = a21 * a33 - a23 * a31;
      const b11 = a22 * a33 - a23 * a32;
      return b00 * b11 - b01 * b10 + b02 * b09 + b03 * b08 - b04 * b07 + b05 * b06;
    }
    throw Error("ArgumentError");
  }
  inverse() {
    const { numRows, numCols } = this;
    if (numRows !== numCols) {
      throw Error("ArgumentError");
    }
    const m = (col, row) => this.valueAt(row, col);
    if (numRows === 2) {
      const det = m(0, 0) * m(1, 1) - m(0, 1) * m(1, 0);
      return new Matrix([
        m(1, 1),
        -m(0, 1),
        -m(1, 0),
        m(0, 0)
      ].map((x) => x / det));
    }
    if (numRows === 3) {
      const a00 = m(0, 0), a01 = m(0, 1), a02 = m(0, 2);
      const a10 = m(1, 0), a11 = m(1, 1), a12 = m(1, 2);
      const a20 = m(2, 0), a21 = m(2, 1), a22 = m(2, 2);
      const b01 = a22 * a11 - a12 * a21;
      const b11 = -a22 * a10 + a12 * a20;
      const b21 = a21 * a10 - a11 * a20;
      const det = a00 * b01 + a01 * b11 + a02 * b21;
      return new Matrix([
        b01,
        -a22 * a01 + a02 * a21,
        a12 * a01 - a02 * a11,
        b11,
        a22 * a00 - a02 * a20,
        -a12 * a00 + a02 * a10,
        b21,
        -a21 * a00 + a01 * a20,
        a11 * a00 - a01 * a10
      ].map((x) => x / det));
    }
    if (numRows === 4) {
      const a00 = m(0, 0), a01 = m(0, 1), a02 = m(0, 2), a03 = m(0, 3);
      const a10 = m(1, 0), a11 = m(1, 1), a12 = m(1, 2), a13 = m(1, 3);
      const a20 = m(2, 0), a21 = m(2, 1), a22 = m(2, 2), a23 = m(2, 3);
      const a30 = m(3, 0), a31 = m(3, 1), a32 = m(3, 2), a33 = m(3, 3);
      const b00 = a00 * a11 - a01 * a10;
      const b01 = a00 * a12 - a02 * a10;
      const b02 = a00 * a13 - a03 * a10;
      const b03 = a01 * a12 - a02 * a11;
      const b04 = a01 * a13 - a03 * a11;
      const b05 = a02 * a13 - a03 * a12;
      const b06 = a20 * a31 - a21 * a30;
      const b07 = a20 * a32 - a22 * a30;
      const b08 = a20 * a33 - a23 * a30;
      const b09 = a21 * a32 - a22 * a31;
      const b10 = a21 * a33 - a23 * a31;
      const b11 = a22 * a33 - a23 * a32;
      const det = b00 * b11 - b01 * b10 + b02 * b09 + b03 * b08 - b04 * b07 + b05 * b06;
      return new Matrix([
        a11 * b11 - a12 * b10 + a13 * b09,
        a02 * b10 - a01 * b11 - a03 * b09,
        a31 * b05 - a32 * b04 + a33 * b03,
        a22 * b04 - a21 * b05 - a23 * b03,
        a12 * b08 - a10 * b11 - a13 * b07,
        a00 * b11 - a02 * b08 + a03 * b07,
        a32 * b02 - a30 * b05 - a33 * b01,
        a20 * b05 - a22 * b02 + a23 * b01,
        a10 * b10 - a11 * b08 + a13 * b06,
        a01 * b08 - a00 * b10 - a03 * b06,
        a30 * b04 - a31 * b02 + a33 * b00,
        a21 * b02 - a20 * b04 - a23 * b00,
        a11 * b07 - a10 * b09 - a12 * b06,
        a00 * b09 - a01 * b07 + a02 * b06,
        a31 * b01 - a30 * b03 - a32 * b00,
        a20 * b03 - a21 * b01 + a22 * b00
      ].map((x) => x / det));
    }
    throw Error("NotImplementedYet");
  }
  isNaN() {
    for (const value of this.values) {
      if (Number.isNaN(value)) {
        return true;
      }
    }
    return false;
  }
  isFinite() {
    for (const value of this.values) {
      if (!Number.isFinite(value)) {
        return false;
      }
    }
    return true;
  }
  toString() {
    const { numRows, numCols, values } = this;
    return `mat${numRows}x${numCols}(${values.join(", ")})`;
  }
};

// src/utils/one-liners.ts
var mix = (a, b, n) => (1 - n) * a + n * b;
var clamp = (x, minVal, maxVal) => Math.min(Math.max(x, minVal), maxVal);

// src/utils/stopwatch.ts
var Stopwatch = class {
  constructor() {
    this._startTime = NaN;
    this._totalElapsed = 0;
    this._running = false;
  }
  start() {
    this._startTime = performance.now();
    this._running = true;
    return this;
  }
  stop() {
    if (this._running) {
      this._running = false;
      this._totalElapsed += performance.now() - this._startTime;
    }
    return this;
  }
  reset() {
    this._totalElapsed = 0;
    this._startTime = this._running ? performance.now() : NaN;
    return this;
  }
  get running() {
    return this._running;
  }
  get elapsedTime() {
    return this._running ? this._totalElapsed + performance.now() - this._startTime : this._totalElapsed;
  }
};

// src/utils/perspective.ts
function ortho(left, right, bottom, top, zNear, zFar) {
  const tx = -(right + left) / (right - left);
  const ty = -(top + bottom) / (top - bottom);
  const tz = -(zFar + zNear) / (zFar - zNear);
  return new Matrix([
    2 / (right - left),
    0,
    0,
    0,
    0,
    2 / (top - bottom),
    0,
    0,
    0,
    0,
    -2 / (zFar - zNear),
    0,
    tx,
    ty,
    tz,
    1
  ]);
}
function frustum(left, right, bottom, top, zNear, zFar) {
  const t1 = 2 * zNear;
  const t2 = right - left;
  const t3 = top - bottom;
  const t4 = zFar - zNear;
  return new Matrix([
    t1 / t2,
    0,
    0,
    0,
    0,
    t1 / t3,
    0,
    0,
    (right + left) / t2,
    (top + bottom) / t3,
    (-zFar - zNear) / t4,
    -1,
    0,
    0,
    -t1 * zFar / t4,
    0
  ]);
}
function perspective(fieldOfView, aspectRatio, zNear, zFar) {
  const y = zNear * Math.tan(fieldOfView * Math.PI / 360);
  const x = y * aspectRatio;
  return frustum(-x, x, -y, y, zNear, zFar);
}

// src/utils/camera.ts
var Camera = class {
  constructor() {
    this.position = new Vector(0, 0, 5);
    this.target = new Vector(0, 0, 0);
    this.up = new Vector(0, 1, 0);
    this.cameraMatrix = mat4_exports.identity();
    this.viewMatrix = mat4_exports.identity();
    this.update();
  }
  update() {
    const { position, target, up } = this;
    const matrix = mat4_exports.lookAt(position, target, up);
    this.cameraMatrix = matrix;
    this.viewMatrix = matrix.inverse();
    return this;
  }
};

// src/shaders/default.vert
var default_default = "precision highp float;attribute vec4 position;attribute vec4 normal;attribute vec2 uv;varying vec4 vPosition;varying vec2 vUv;varying vec4 vNormal;void main(){vUv=uv;vPosition=position;vNormal=normal;gl_Position=position;}";

// src/shaders/default.frag
var default_default2 = "precision highp float;void main(){gl_FragColor=vec4(1.,0.,0.,1.);}";

// src/shaders/basic.frag
var basic_default = "precision highp float;uniform vec4 color;void main(){gl_FragColor=color;}";

// src/shaders/normal.frag
var normal_default = "precision highp float;varying vec4 vPosition;varying vec4 vNormal;varying vec2 vUv;varying float vFaceIndex;uniform float time;uniform vec2 resolution;void main(){vec3 color=vec3(vNormal.x,vNormal.y,1.);gl_FragColor=vec4(color,1.);}";

// src/webgl/material.ts
function createDefaultMaterial() {
  return {
    vertexShader: default_default,
    fragmentShader: default_default2,
    drawMode: WebGLRenderingContext.TRIANGLES,
    uniforms: {}
  };
}
function createNormalMaterial() {
  return {
    vertexShader: default_default,
    fragmentShader: normal_default,
    drawMode: WebGLRenderingContext.TRIANGLES,
    uniforms: {}
  };
}
function createBasicMaterial(color = "#ff0000") {
  return {
    vertexShader: default_default,
    fragmentShader: basic_default,
    drawMode: WebGLRenderingContext.TRIANGLES,
    uniforms: {
      color: Color.fromHex(color)
    }
  };
}
function createShaderMaterial(vertexShader = default_default, fragmentShader = default_default2, uniforms = {}, drawMode = WebGLRenderingContext.TRIANGLES) {
  return {
    vertexShader,
    fragmentShader,
    drawMode,
    uniforms
  };
}

// src/webgl/texture.ts
var ERRORS = {
  LOADING_ERROR: "Error loading image"
};
var Wrapping = {
  CLAMP_TO_EDGE: WebGLRenderingContext.CLAMP_TO_EDGE,
  REPEAT: WebGLRenderingContext.REPEAT,
  MIRRORED_REPEAT: WebGLRenderingContext.MIRRORED_REPEAT
};
var Filter = {
  LINEAR: WebGLRenderingContext.LINEAR,
  NEAREST: WebGLRenderingContext.NEAREST,
  LINEAR_MIPMAP_LINEAR: WebGLRenderingContext.LINEAR_MIPMAP_LINEAR,
  LINEAR_MIPMAP_NEAREST: WebGLRenderingContext.LINEAR_MIPMAP_NEAREST,
  NEAREST_MIPMAP_LINEAR: WebGLRenderingContext.NEAREST_MIPMAP_LINEAR,
  NEAREST_MIPMAP_NEAREST: WebGLRenderingContext.NEAREST_MIPMAP_NEAREST
};
var Texture = class {
  constructor(image, options) {
    this.minFilter = Filter.NEAREST;
    this.magFilter = Filter.NEAREST;
    this.wrapS = Wrapping.CLAMP_TO_EDGE;
    this.wrapT = Wrapping.CLAMP_TO_EDGE;
    this.textureIndex = 0;
    this.gl = null;
    this.texture = null;
    this.image = image;
    if (options) {
      Object.assign(this, options);
    }
  }
  static async fromImageUrl(url, options) {
    return new Promise((resolve, reject) => {
      const image = new Image();
      image.crossOrigin = "anonymous";
      image.src = url;
      image.onload = () => {
        resolve(new Texture(image, options));
      };
      image.onerror = () => {
        reject(Error(ERRORS.LOADING_ERROR));
      };
    });
  }
  upload(gl, textureIndex) {
    if (this.gl) {
      this.dispose();
    }
    const texture = gl.createTexture();
    const idx = typeof textureIndex !== "undefined" ? textureIndex : this.textureIndex;
    this.textureIndex = idx;
    gl.activeTexture(gl.TEXTURE0 + idx);
    gl.bindTexture(gl.TEXTURE_2D, texture);
    gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_MIN_FILTER, this.minFilter);
    gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_MAG_FILTER, this.magFilter);
    gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_WRAP_S, this.wrapS);
    gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_WRAP_T, this.wrapT);
    gl.texImage2D(gl.TEXTURE_2D, 0, gl.RGBA, gl.RGBA, gl.UNSIGNED_BYTE, this.image);
    this.gl = gl;
    this.texture = texture;
    return this;
  }
  update() {
    const { gl, texture, textureIndex } = this;
    if (!gl) {
      throw Error();
    }
    gl.activeTexture(gl.TEXTURE0 + textureIndex);
    gl.bindTexture(gl.TEXTURE_2D, texture);
    gl.texSubImage2D(gl.TEXTURE_2D, 0, 0, 0, gl.RGBA, gl.UNSIGNED_BYTE, this.image);
    return this;
  }
  dispose() {
    if (this.texture && this.gl) {
      this.gl.deleteTexture(this.texture);
      this.texture = null;
      this.gl = null;
    }
  }
};

// src/utils/uniform-helpers.ts
function setUniformVector(gl, loc, value) {
  switch (value.length) {
    case 1:
      gl.uniform1fv(loc, value);
      return true;
    case 2:
      gl.uniform2fv(loc, value);
      return true;
    case 3:
      gl.uniform3fv(loc, value);
      return true;
    case 4:
      gl.uniform4fv(loc, value);
      return true;
  }
  return false;
}
function setUniformIntVector(gl, loc, value) {
  switch (value.length) {
    case 1:
      gl.uniform1iv(loc, value);
      return true;
    case 2:
      gl.uniform2iv(loc, value);
      return true;
    case 3:
      gl.uniform3iv(loc, value);
      return true;
    case 4:
      gl.uniform4iv(loc, value);
      return true;
  }
  return false;
}
function setUniformMatrix(gl, loc, value, transpose = false) {
  if (value.length === 4) {
    gl.uniformMatrix2fv(loc, transpose, value);
    return true;
  }
  if (value.length === 9) {
    gl.uniformMatrix3fv(loc, transpose, value);
    return true;
  }
  if (value.length === 16) {
    gl.uniformMatrix4fv(loc, transpose, value);
    return true;
  }
  return false;
}
function setUniform(gl, program, name, uniform) {
  gl.useProgram(program);
  const loc = gl.getUniformLocation(program, name);
  if (typeof uniform === "number") {
    gl.uniform1f(loc, uniform);
    return true;
  }
  if (uniform instanceof Texture) {
    gl.uniform1i(loc, uniform.textureIndex);
    return true;
  }
  if (uniform instanceof Vector) {
    return setUniformIntVector(gl, loc, uniform.values);
  }
  if (uniform instanceof Matrix) {
    return setUniformMatrix(gl, loc, uniform.values);
  }
  if (uniform instanceof Color) {
    return setUniformVector(gl, loc, uniform.toVec4());
  }
  if (uniform instanceof Array) {
    if (typeof uniform[0] === "number") {
      return setUniformVector(gl, loc, uniform);
    }
    if (typeof uniform[0] === "bigint") {
      const converter = (val) => parseInt(val.toString(), 10);
      return setUniformIntVector(gl, loc, uniform.map(converter));
    }
    if (uniform[0] instanceof Array) {
      return setUniformMatrix(gl, loc, uniform.flat(1));
    }
  }
  return false;
}
function wrapUniforms(gl, program, uniform) {
  const handler = {
    get(target, prop) {
      return target[prop];
    },
    set(target, prop, value) {
      target[prop] = value;
      return setUniform(gl, program, prop, value);
    }
  };
  return new Proxy(uniform, handler);
}

// src/utils/indexed-geometry.ts
function createIndexedGeometry(geometry) {
  if (geometry.index !== null) {
    throw Error("the buffer geometry is already indexed");
  }
  const result = new BufferGeometry();
  const newAttribs = {};
  const indices = [];
  let lastIndex = -1;
  const indexMap = {};
  for (let i = 0; i < geometry.count; i++) {
    const data = {};
    for (const [key, attribute] of Object.entries(geometry.attributes)) {
      const offset = i * attribute.recordSize;
      data[key] = Array.from(attribute.data.slice(offset, offset + attribute.recordSize));
    }
    const serialized = JSON.stringify(data);
    if (indexMap[serialized]) {
      const index = indexMap[serialized];
      indices.push(index);
    } else {
      lastIndex++;
      indexMap[serialized] = lastIndex;
      indices.push(lastIndex);
      for (const [key, value] of Object.entries(data)) {
        if (!newAttribs[key]) {
          newAttribs[key] = [];
        }
        Array.prototype.push.apply(newAttribs[key], value);
      }
    }
  }
  result.setIndex(indices, lastIndex > 65534 ? 32 : 16);
  for (const [key, value] of Object.entries(newAttribs)) {
    const recordSize = geometry.attributes[key].recordSize;
    result.setAttribute(key, new BufferAttribute(new Float32Array(value), recordSize));
  }
  return result;
}

// src/utils/merge-geometries.ts
function mergeNonIndexedGeometries(a, b) {
  const countA = a.index !== null ? a.count : Math.max.apply(null, Object.values(a.attributes).map((attr) => attr.data.length / attr.recordSize | 0));
  const countB = b.index !== null ? b.count : Math.max.apply(null, Object.values(b.attributes).map((attr) => attr.data.length / attr.recordSize | 0));
  const result = new BufferGeometry();
  for (const key of Object.keys(a.attributes)) {
    const newAttrib = [];
    const attribA = a.attributes[key];
    const attribB = b.attributes[key] || new BufferAttribute(new Float32Array(a.count), 1);
    const recordSize = Math.max(attribA.recordSize, (attribB == null ? void 0 : attribB.recordSize) || 0);
    for (let i = 0; i < countA; i++) {
      for (let j = 0; j < recordSize; j++) {
        if (attribA.recordSize <= j || attribA.data.length <= i * attribA.recordSize + j) {
          newAttrib.push(0);
          continue;
        }
        newAttrib.push(a.attributes[key].data[i * attribA.recordSize + j]);
      }
    }
    for (let i = 0; i < countB; i++) {
      for (let j = 0; j < recordSize; j++) {
        if (!attribB || attribB.recordSize <= j || attribB.data.length <= i * attribB.recordSize + j) {
          newAttrib.push(0);
          continue;
        }
        newAttrib.push(b.attributes[key].data[i * attribB.recordSize + j]);
      }
    }
    result.setAttribute(key, new BufferAttribute(new Float32Array(newAttrib), recordSize));
  }
  if (a.groups.length === 0) {
    result.groups.push({ startIndex: 0, count: countA });
  }
  result.groups.push(...a.groups, {
    startIndex: countA,
    count: countB
  });
  return result;
}
function mergeIndexedGeometries(a, b) {
  if (a.index === null || b.index === null) {
    throw Error();
  }
  const result = mergeNonIndexedGeometries(a, b);
  const maxIndexA = Math.max.apply(null, a.index);
  const maxIndexB = Math.max.apply(null, b.index);
  const indices = a.index.concat(b.index.map((idx) => maxIndexA + 1 + idx));
  result.setIndex(indices, maxIndexA + maxIndexB + 1 < 65534 ? 16 : 32);
  return result;
}
function mergeGeometries(geometry, otherGeometry) {
  if (geometry.index === null && otherGeometry.index === null) {
    return mergeNonIndexedGeometries(geometry, otherGeometry);
  }
  return mergeIndexedGeometries(geometry.index === null ? createIndexedGeometry(geometry) : geometry, otherGeometry.index === null ? createIndexedGeometry(otherGeometry) : otherGeometry);
}

// src/utils/mat2.ts
var mat2_exports = {};
__export(mat2_exports, {
  rotation: () => rotation,
  scaling: () => scaling
});
function rotation(angle) {
  const S = Math.sin(angle);
  const C = Math.cos(angle);
  return new Matrix([
    C,
    S,
    -S,
    C
  ]);
}
function scaling(sx, sy) {
  return new Matrix([
    sx,
    0,
    0,
    sy
  ]);
}

// src/utils/mat3.ts
var mat3_exports = {};
__export(mat3_exports, {
  rotX: () => rotX,
  rotY: () => rotY,
  rotZ: () => rotZ,
  scaling: () => scaling2,
  translation: () => translation
});
function translation(x, y) {
  return new Matrix([
    1,
    0,
    0,
    0,
    1,
    0,
    x,
    y,
    1
  ]);
}
function scaling2(sx, sy, sz) {
  return new Matrix([
    sx,
    0,
    0,
    0,
    sy,
    0,
    0,
    0,
    sz
  ]);
}
function rotX(angle) {
  const { sin, cos } = Math;
  const S = sin(angle);
  const C = cos(angle);
  return new Matrix([
    1,
    0,
    0,
    0,
    C,
    S,
    0,
    -S,
    C
  ]);
}
function rotY(angle) {
  const { sin, cos } = Math;
  const S = sin(angle);
  const C = cos(angle);
  return new Matrix([
    C,
    0,
    -S,
    0,
    1,
    0,
    S,
    0,
    C
  ]);
}
function rotZ(angle) {
  const { sin, cos } = Math;
  const S = sin(angle);
  const C = cos(angle);
  return new Matrix([
    C,
    S,
    0,
    -S,
    C,
    0,
    0,
    0,
    1
  ]);
}

// src/utils/mat4.ts
var mat4_exports = {};
__export(mat4_exports, {
  identity: () => identity,
  lookAt: () => lookAt,
  rotX: () => rotX2,
  rotY: () => rotY2,
  rotZ: () => rotZ2,
  scaling: () => scaling3,
  translation: () => translation2
});
function identity() {
  return new Matrix([
    1,
    0,
    0,
    0,
    0,
    1,
    0,
    0,
    0,
    0,
    1,
    0,
    0,
    0,
    0,
    1
  ]);
}
function translation2(x, y, z) {
  return new Matrix([
    1,
    0,
    0,
    0,
    0,
    1,
    0,
    0,
    0,
    0,
    1,
    0,
    x,
    y,
    z,
    1
  ]);
}
function scaling3(sx, sy, sz) {
  return new Matrix([
    sx,
    0,
    0,
    0,
    0,
    sy,
    0,
    0,
    0,
    0,
    sz,
    0,
    0,
    0,
    0,
    1
  ]);
}
function rotX2(angle) {
  const { sin, cos } = Math;
  const S = sin(angle);
  const C = cos(angle);
  return new Matrix([
    1,
    0,
    0,
    0,
    0,
    C,
    S,
    0,
    0,
    -S,
    C,
    0,
    0,
    0,
    0,
    1
  ]);
}
function rotY2(angle) {
  const { sin, cos } = Math;
  const S = sin(angle);
  const C = cos(angle);
  return new Matrix([
    C,
    0,
    -S,
    0,
    0,
    1,
    0,
    0,
    S,
    0,
    C,
    0,
    0,
    0,
    0,
    1
  ]);
}
function rotZ2(angle) {
  const { sin, cos } = Math;
  const S = sin(angle);
  const C = cos(angle);
  return new Matrix([
    C,
    S,
    0,
    0,
    -S,
    C,
    0,
    0,
    0,
    0,
    1,
    0,
    0,
    0,
    0,
    1
  ]);
}
function lookAt(position, target, up) {
  const zAxis = position.sub(target).normalized;
  const xAxis = up.cross(zAxis).normalized;
  const yAxis = zAxis.cross(xAxis).normalized;
  return new Matrix([
    xAxis.x,
    xAxis.y,
    xAxis.z,
    0,
    yAxis.x,
    yAxis.y,
    yAxis.z,
    0,
    zAxis.x,
    zAxis.y,
    zAxis.z,
    0,
    position.x,
    position.y,
    position.z,
    1
  ]);
}

// src/geometries/plane.ts
function createPlaneGeometry(width = 2, height = 2, rows = 1, cols = 1) {
  const cols1 = cols + 1;
  const rows1 = rows + 1;
  const deltaX = width / cols;
  const deltaY = height / rows;
  const vertices = Array(cols1 * rows1).fill(0).map((_, i) => {
    const ix = i % cols1;
    const iy = i / cols1 | 0;
    return new Vector(ix * deltaX - width / 2, iy * deltaY - height / 2, 0);
  });
  const uvs = Array(cols1 * rows1).fill(0).map((_, i) => {
    const ix = i % cols1;
    const iy = i / cols1 | 0;
    return new Vector(ix / cols, 1 - iy / rows);
  });
  const faces = Array(rows * cols).fill(0).map((_, i) => {
    const ix = i % cols;
    const iy = i / cols | 0;
    const idx = iy * rows1 + ix;
    return [
      [idx, idx + 1, idx + rows1],
      [idx + 1, idx + rows1 + 1, idx + rows1]
    ];
  }).flat(1);
  const normals = new Float32Array(faces.map(() => {
    const n = [0, 0, 1];
    return [n, n, n].flat(1);
  }).flat(1));
  const geometry = new BufferGeometry();
  const positionBuffer = new Float32Array(facesToBuffer(faces, vertices));
  const uvBuffer = new Float32Array(facesToBuffer(faces, uvs));
  geometry.setAttribute("position", new BufferAttribute(positionBuffer, 3));
  geometry.setAttribute("normal", new BufferAttribute(normals, 3));
  geometry.setAttribute("uv", new BufferAttribute(uvBuffer, 2));
  return geometry;
}

// src/geometries/sphere.ts
function createSphereGeometry(r = 0.5, sides = 36, segments = 18) {
  const vertices = [];
  const normals = [];
  const texCoords = [];
  const faces = [];
  const dphi = 360 / sides;
  const dtheta = 180 / segments;
  const evalPos = (theta, phi) => {
    const deg = Math.PI / 180;
    const pos = new Vector(r * Math.sin(theta * deg) * Math.sin(phi * deg), r * Math.cos(theta * deg), r * Math.sin(theta * deg) * Math.cos(phi * deg));
    return pos;
  };
  for (let segment = 0; segment <= segments; segment++) {
    const theta = segment * dtheta;
    for (let side = 0; side <= sides; side++) {
      const phi = side * dphi;
      const pos = evalPos(theta, phi);
      const texCoord = new Vector(phi / 360, theta / 180);
      const normal = pos.normalized;
      vertices.push(pos);
      texCoords.push(texCoord);
      normals.push(normal);
      if (segment === segments)
        continue;
      if (side === sides)
        continue;
      if (segment == 0) {
        faces.push([
          segment * (sides + 1) + side,
          (segment + 1) * (sides + 1) + side,
          (segment + 1) * (sides + 1) + side + 1
        ]);
      } else if (segment == segments - 1) {
        faces.push([
          segment * (sides + 1) + side,
          (segment + 1) * (sides + 1) + side + 1,
          segment * (sides + 1) + side + 1
        ]);
      } else {
        const A = segment * (sides + 1) + side;
        const B = (segment + 1) * (sides + 1) + side;
        const C = (segment + 1) * (sides + 1) + side + 1;
        const D = segment * (sides + 1) + side + 1;
        faces.push([A, B, D]);
        faces.push([B, C, D]);
      }
    }
  }
  const geometry = new BufferGeometry();
  const positionData = new Float32Array(facesToBuffer(faces, vertices));
  const normalData = new Float32Array(facesToBuffer(faces, normals));
  const uvData = new Float32Array(facesToBuffer(faces, texCoords));
  geometry.setAttribute("position", new BufferAttribute(positionData, 3));
  geometry.setAttribute("normal", new BufferAttribute(normalData, 3));
  geometry.setAttribute("uv", new BufferAttribute(uvData, 2));
  return geometry;
}

// src/webgl/webgl-errors.ts
var ERRORS2 = {
  WEBGL_INIT: "WebGL initialization failed",
  SHADER_FAIL: "Shader compile failed",
  WEBGL_ERROR: "WebGL error"
};

// src/webgl/mesh.ts
var Mesh = class {
  constructor(geometry, material, bufferUsage = WebGLRenderingContext.DYNAMIC_DRAW) {
    this.geometry = geometry;
    this.material = material;
    this.bufferUsage = bufferUsage;
    this.buffers = {};
    this.program = null;
    this.gl = null;
  }
  init(gl) {
    if (this.gl) {
      return;
    }
    this.gl = gl;
    const { material } = this;
    const { fragmentShader, vertexShader } = material;
    this.program = this.createProgram(vertexShader, fragmentShader);
    const uniformsProxy = wrapUniforms(this.gl, this.program, material.uniforms);
    material.uniforms = uniformsProxy;
    this.createBuffers();
    gl.useProgram(this.program);
    this.setUniforms();
  }
  compileShader(type, code) {
    const { gl } = this;
    if (!gl) {
      throw Error(ERRORS2.WEBGL_INIT);
    }
    const sh = gl.createShader(type);
    if (!sh) {
      throw Error("error creating shader.");
    }
    gl.shaderSource(sh, code);
    gl.compileShader(sh);
    if (!gl.getShaderParameter(sh, gl.COMPILE_STATUS)) {
      throw gl.getShaderInfoLog(sh);
    }
    return sh;
  }
  createProgram(vertexShader, fragmentShader) {
    const { gl } = this;
    if (!gl) {
      throw Error("WebGL context not initialized");
    }
    const program = gl.createProgram();
    if (!program) {
      throw Error(ERRORS2.SHADER_FAIL);
    }
    const fragShader = this.compileShader(gl.FRAGMENT_SHADER, fragmentShader) || null;
    const vertShader = this.compileShader(gl.VERTEX_SHADER, vertexShader) || null;
    if (!fragShader || !vertShader) {
      throw Error(ERRORS2.SHADER_FAIL);
    }
    gl.attachShader(program, fragShader);
    gl.attachShader(program, vertShader);
    gl.linkProgram(program);
    if (!gl.getProgramParameter(program, gl.LINK_STATUS)) {
      throw gl.getProgramInfoLog(program);
    }
    return program;
  }
  createBuffers() {
    const { gl, program, geometry } = this;
    if (!gl || !program) {
      throw Error(ERRORS2.WEBGL_INIT);
    }
    const buffers = {};
    for (const [name, attrib] of Object.entries(geometry.attributes)) {
      const buffer = gl.createBuffer();
      if (!buffer) {
        throw Error(ERRORS2.WEBGL_INIT);
      }
      const attribLoc = gl.getAttribLocation(program, name);
      if (attribLoc === -1) {
        continue;
      }
      gl.bindBuffer(gl.ARRAY_BUFFER, buffer);
      gl.bufferData(gl.ARRAY_BUFFER, attrib.data, this.bufferUsage);
      gl.enableVertexAttribArray(attribLoc);
      gl.vertexAttribPointer(attribLoc, attrib.recordSize, gl.FLOAT, false, 0, 0);
      buffers[name] = buffer;
    }
    if (geometry.index) {
      const indexBuffer = gl.createBuffer();
      if (indexBuffer === null) {
        throw Error(ERRORS2.WEBGL_ERROR);
      }
      gl.bindBuffer(gl.ELEMENT_ARRAY_BUFFER, indexBuffer);
      let data = null;
      if (geometry.indexType === 16) {
        data = new Uint16Array(geometry.index);
      }
      if (geometry.indexType === 32) {
        data = new Uint32Array(geometry.index);
      }
      gl.bufferData(gl.ELEMENT_ARRAY_BUFFER, data, this.bufferUsage);
      buffers["_index"] = indexBuffer;
    }
    this.buffers = buffers;
    return this;
  }
  deleteBuffers() {
    const { gl, program } = this;
    if (!gl || !program) {
      return this;
    }
    for (const [key, buffer] of Object.entries(this.buffers)) {
      if (!key.startsWith("_")) {
        const loc = gl.getAttribLocation(program, key);
        gl.disableVertexAttribArray(loc);
      }
      gl.deleteBuffer(buffer);
    }
    this.buffers = {};
    return this;
  }
  updateBuffers() {
    if (!this.gl || !this.program) {
      throw Error(ERRORS2.WEBGL_INIT);
    }
    this.deleteBuffers();
    this.createBuffers();
    return this;
  }
  updateBuffer(attribName, newData, offset = 0, syncGeometry = true) {
    const { geometry, gl } = this;
    const buffer = this.buffers[attribName];
    if (syncGeometry) {
      geometry.attributes[attribName].data.set(newData, offset);
    }
    if (gl && buffer !== null) {
      gl.bindBuffer(gl.ARRAY_BUFFER, buffer);
      gl.bufferSubData(gl.ARRAY_BUFFER, offset, newData);
    }
    return this;
  }
  recompile() {
    const { gl, program } = this;
    const { vertexShader, fragmentShader } = this.material;
    if (gl && program) {
      gl.deleteProgram(program);
    }
    this.program = this.createProgram(vertexShader, fragmentShader);
    return this;
  }
  setUniforms() {
    const { gl, program, material } = this;
    if (!gl || !program) {
      throw Error(ERRORS2.WEBGL_INIT);
    }
    gl.useProgram(program);
    for (const [name, value] of Object.entries(material.uniforms)) {
      setUniform(gl, program, name, value);
    }
    return this;
  }
  disableAttribs() {
    const { gl, program, buffers } = this;
    if (!gl || !program) {
      throw Error(ERRORS2.WEBGL_INIT);
    }
    gl.useProgram(program);
    for (const key of Object.keys(buffers)) {
      const loc = gl.getAttribLocation(program, key);
      if (loc > -1) {
        gl.disableVertexAttribArray(loc);
      }
    }
    return this;
  }
  enableAttribs() {
    const { gl, program, buffers, geometry } = this;
    if (!gl || !program) {
      throw Error(ERRORS2.WEBGL_INIT);
    }
    gl.useProgram(program);
    for (const [key, buffer] of Object.entries(buffers)) {
      const attribute = geometry.attributes[key];
      const loc = gl.getAttribLocation(program, key);
      if (loc > -1) {
        gl.enableVertexAttribArray(loc);
        gl.bindBuffer(gl.ARRAY_BUFFER, buffer);
        gl.vertexAttribPointer(loc, attribute.recordSize, gl.FLOAT, false, 0, 0);
      }
    }
    return this;
  }
  draw() {
    const { gl, geometry, material } = this;
    if (!gl) {
      throw Error(ERRORS2.WEBGL_INIT);
    }
    if (geometry.index !== null) {
      let indexType = WebGLRenderingContext.NONE;
      if (geometry.indexType === 16) {
        indexType = WebGLRenderingContext.UNSIGNED_SHORT;
      }
      if (geometry.indexType === 32) {
        indexType = WebGLRenderingContext.UNSIGNED_INT;
      }
      gl.drawElements(material.drawMode, geometry.count, indexType, 0);
      return this;
    }
    gl.drawArrays(material.drawMode, 0, geometry.count);
    return this;
  }
  dispose() {
    const { gl, program } = this;
    if (!gl || !program) {
      return;
    }
    this.deleteBuffers();
    gl.deleteProgram(program);
    this.gl = null;
    this.program = null;
    this.material.uniforms = { ...this.material.uniforms };
  }
};

// src/webgl/renderer.ts
var Renderer = class {
  constructor(canvas) {
    this.canvas = canvas;
    this.pixelRatio = 1;
    const gl = this.canvas.getContext("webgl");
    if (!gl) {
      throw Error(ERRORS2.WEBGL_INIT);
    }
    this.gl = gl;
  }
  render(scene) {
    for (const obj of scene) {
      obj.init(this.gl);
      obj.enableAttribs();
      obj.draw();
      obj.disableAttribs();
    }
    return this;
  }
  setPixelRatio(pixelRatio) {
    this.pixelRatio = pixelRatio;
    return this;
  }
  setSize(width, height) {
    const { gl, canvas, pixelRatio } = this;
    canvas.width = width * pixelRatio;
    canvas.height = height * pixelRatio;
    gl.viewport(0, 0, gl.drawingBufferWidth, gl.drawingBufferHeight);
    return this;
  }
  dispose() {
    const { gl } = this;
    const loseCtx = gl.getExtension("WEBGL_lose_context");
    if (loseCtx && typeof loseCtx.loseContext === "function") {
      loseCtx.loseContext();
    }
  }
};
export {
  BufferAttribute,
  BufferGeometry,
  Camera,
  Color,
  Filter,
  mat2_exports as Mat2,
  mat3_exports as Mat3,
  mat4_exports as Mat4,
  Matrix,
  Mesh,
  Renderer,
  Stopwatch,
  Texture,
  Vector,
  Wrapping,
  calculateSurfaceNormal,
  clamp,
  createBasicMaterial,
  createBoxGeometry,
  createDefaultMaterial,
  createIndexedGeometry,
  createNormalMaterial,
  createPlaneGeometry,
  createShaderMaterial,
  createSphereGeometry,
  facesToBuffer,
  frustum,
  mergeGeometries,
  mix,
  ortho,
  perspective,
  setUniform,
  wrapUniforms
};
