import { Vector } from './vector';
/** @class Matrix */
export declare class Matrix {
    values: number[];
    numRows: number;
    numCols: number;
    constructor(values: number[], options?: {
        numRows: number;
        numCols: number;
    });
    /**
     * create identity matrix
     * @param dimension dimension of the matrix
     */
    static identity(dimension: number): Matrix;
    /**
     * Converts a vector with dimension n into a matrix with 1 col and n rows
     * useful for matrix multiplication
     * @param value the input vector
     */
    static fromVector(value: Vector): Matrix;
    /**
     * Converts a bunch of vectors into a matrix
     */
    static fromVectors(vectors: Vector[]): Matrix;
    /**
     * convert to array
     */
    toArray(): number[];
    /**
     * get value at a given position
     * @param row row index starting from 0
     * @param column column index starting from 0
     */
    valueAt(row: number, column: number): number;
    /**
     * get column at a given index
     * @param column index of column starting from 0
     * @returns column as an array of numbers
     */
    colAt(column: number): number[];
    /**
     * get row at a given index
     * @param row index of row starting from 0
     * @returns row as an array of numbers
     */
    rowAt(row: number): number[];
    /**
     * returns transposed matrix
     */
    transpose(): Matrix;
    /**
     * check for equality
     * @param otherMatrix matrix to compare
     * @returns true or false
     */
    equals(otherMatrix: Matrix): boolean;
    /**
     * check for rough equality (because in JS, .1+.2 !== .3)
     * @param otherMatrix matrix to compare
     * @param delta tolerance, by default 1e-14
     * @returns true or false
     */
    roughlyEquals(otherMatrix: Matrix, delta?: number): boolean;
    /**
     * add two matrices
     * @param otherMatrix matrix to add
     * @returns result matrix
     */
    add(otherMatrix: Matrix): Matrix;
    /**
     * subtract another matrix
     * @param otherMatrix matrix to subtract
     * @returns result matrix
     */
    sub(otherMatrix: Matrix): Matrix;
    /**
     * matrix multiplication
     * @param param can be a matrix, a number or a vector
     * @returns a vector in case of matrix vector multiplication, else a matrix
     * @throws dimension Mismatch if dimensions doesn't match
     */
    mul(param: Matrix | number | Vector): Matrix | Vector;
    div(param: number | Matrix): Matrix;
    /**
     * calculate determinant, implemented for mat2, mat3, mat4
     */
    determinant(): number;
    /**
     * calculate M^-1 for mat2, mat3, mat4
     */
    inverse(): Matrix;
    /**
     * Checks if the matrix contains NaN values
     * @returns true if one of the values is NaN
     */
    isNaN(): boolean;
    /**
     * Checks if the matrix contains NaN or Infinity values
     * @returns true if all values are finite (neither NaN nor Infinity)
     */
    isFinite(): boolean;
    /**
     * convert matrix to string
     * @returns a string containing matROWSxCOLS(comma-separated-values)
     */
    toString(): string;
}
